//
// RIMVersion.h
//
// Module dependency marking macros for RIM handheld device
// applications built with Microsoft development tools.
//
// General Usage
// =============
//
// Invoke the macro RIM_DEFINE_VERSION to mark
// an executable module (DLL, usually) as supplying
// a particular version of an interface, that is
// binary compatible with a previous interface. A single 
// RIMversion structure instance is emitted in the ".rimdep"
// section of the module for each unique name the macro
// is invoked with.
//
// Invoke the macro RIM_DEFINE_DEPENDENCY to mark
// an executable module as depending on a particular
// version of an interface.  A single RIMDependency structure
// instance will be emitted in the ".rimdep" section of
// the module for each unique name the macro is invoked
// with.
//
// For both macros, invoking them multiple times with
// identical arguments from different object modules
// is fine. Invoking them with the same name, but different
// version numbers will cause only one of the version
// records to be retained.
//
// ie. this is OK:
//
//     RIM_DEFINE_DEPENDENCY(Foo, 1, 5)		// (in file1.cpp)
//     RIM_DEFINE_DEPENDENCY(Foo, 1, 5)		// (in file2.cpp)
//
// But this isn't (version info will be one of the two,
// but which one isn't readily predictable):
//
//     RIM_DEFINE_DEPENDENCY(Foo, 1, 5)		// (in file1.cpp)
//     RIM_DEFINE_DEPENDENCY(Foo, 2, 0)		// (in file2.cpp)
//
//
// Suggested Usage Sketch
// ======================
//
// In the principal public interface header(s) of your module,
// add code along the lines of:
//
// #include "RIMVersion.h"
//
// #define FOO_NAME "Foo"
// #define FOO_MAJOR    1
// #define FOO_MINOR    2
// #define FOO_RELEASE  1
// #define FOO_BUILD    4
// #define FOO_MIN_COMPAT_MAJOR 1	// Foo 1.2 is binary compatible with Foo 1.0
// #define FOO_MIN_COMPAT_MINOR 0
//
// #ifdef IMPLEMENTING_FOO
// // Define a version record, with the semantics:
// // "This is module "Foo" at version 1.2.1.4, and it is backwards binary
// //  compatible back to 1.0.x.x"
// RIM_DEFINE_VERSION(FOO_NAME, FOO_MAJOR, FOO_MINOR, FOO_RELEASE, FOO_BUILD, \ 
//                    FOO_MIN_COMPAT_MAJOR, FOO_MIN_COMPAT_MINOR)
// #else
// // Define a dependency record, with the semantics:
// //   "This module requires a version of "Foo" exposing an interface
// //    that is binary compatible with Foo 1.0"
// RIM_DEFINE_DEPENDENCY(FOO_NAME, FOO_MIN_COMPAT_MAJOR, FOO_MIN_COMPAT_MINOR)
// #endif
// 

#if 0
#pragma once
#endif

#ifndef RIMVERSION_H_included
#define RIMVERSION_H_included

#pragma pack(push, 1)

struct RIMVersion {
	const struct RIMVersion * thisRec; // Self-referential fingerprint.
	unsigned char	size;		// Size of this record (including name data)
	unsigned char	flags;		// Flags (reserved; must be 0x00)
	unsigned short	verMajor;	// Major version number of this module
	unsigned short	verMinor;	// Minor version number of this module
	unsigned short	verPatch;	// Patch version number of this module
	unsigned short	verBuild;	// Build number of this module
	unsigned short	minCompatMajor;	// Minimum binary compatible major version
	unsigned short	minCompatMinor;	// Minimum binary compatible minor version
	char			name[];		// This module's name.
};

struct RIMDependency {
	const struct RIMDependency * thisRec; // Self-referential fingerprint.
	unsigned char   size;		// Size of this record (including name data)
	unsigned char   flags;		// Flags (reserved: must be 0x01)
	unsigned short  verReqMajor;// Major version of module this depends on
	unsigned short  verReqMinor;// Minor version of module this depends on
	char			name[];		// Module that this record shows a dependency on
};

#pragma pack(pop)

#if 0
// Section for module version definition records.
#define RIMDEPENDENCY_SECTION   ".rimdep"

// Ensure the segments are known to the compiler,
// for the sake of __declspec(allocate())
//#pragma const_seg(RIMDEPENDENCY_SECTION)
//#pragma const_seg()
#pragma data_seg(".rimdep")
#pragma data_seg()

// Mark the section as read-only.
#pragma comment(linker, "/SECTION:.rimdep,R")

// Create the symbol name for a version record.
#define _RIMVERNAME(x)	_v##x

// Create the symbol name for a dependency record.
#define _RIMDEPNAME(x)	_d##x

#define _RIMQUOTE(x)	   #x

// Define a RIMVersion instance in the RIMDEPENDENCY_SECTION section.
#define RIM_DEFINE_VERSION(name,major,minor,patch,build,compMaj,compMin) \
 __declspec(dllexport) \
	struct RIMVersion _RIMVERNAME(name) = \
		{ &_RIMVERNAME(name), \
		  sizeof(struct RIMVersion) + sizeof(_RIMQUOTE(name)), \
	   	  0x00, major, minor, patch, build, compMaj, compMin, _RIMQUOTE(name) };

// Define a RIMDependency instance in the RIMDEPENDENCY_SECTION section.
#define RIM_DEFINE_DEPENDENCY(name,reqMajor,reqMinor) \
 __declspec(dllexport) \
	struct RIMDependency _RIMDEPNAME(name) = \
		{ &_RIMDEPNAME(name), \
		  sizeof(struct RIMDependency) + sizeof(_RIMQUOTE(name)), \
		  0x01, reqMajor, reqMinor, _RIMQUOTE(name) };
#else
#define RIM_DEFINE_VERSION(name, major, minor, patch, build, compMaj, compMin) 
#define RIM_DEFINE_DEPENDENCY(name, reqMajor, reqMinor)
#endif

#endif /* RIMVERSION_H_included */
